package parametriche;

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.List;
import java.util.NoSuchElementException;

/**
* Classe che realizza una collezione appoggiandosi ad un'array.
* Non Permette elementi duplicati. Permette null.
*/
public class ArrayCollection<E>
implements Collection<E>
{
	E[] storage = null;
	int count;

	long when = -1;
	int who = this.hashCode();

	public ArrayCollection(int dimMax)
	{
		storage = (E[])(new Object[dimMax]);
		count = 0;
	}

	public ArrayCollection()
	{
		this(10);
	}

	/**
	* Metodo di inserimento alla fine della collezione.
	* Non permette elementi duplicati. Accetta null.
	* Ritorna true se la collezione e' modificata dall'operazione
	* corrente.
	*/
	public boolean add(E elem)
	{
		//dummy implementation secondo la convenzione delle optional operations
		//throw new UnsupportedOperationException("torna domani");
		if(elem == null)
		{
			for(int i = 0;i<count;i++)
			{
				if(storage[i] == null)
				{
					return false;
				}
			}
		}
		else
		{
			for(int i = 0;i<count;i++)
			{
				if(elem.equals(storage[i]))
				{
					return false;
				}
			}
		}
		try
		{
			storage[count++] = elem;
		}
		catch(ArrayIndexOutOfBoundsException aioobe)
		{
			IllegalStateException ise = new IllegalStateException("Collection full");
			ise.initCause(aioobe);
			throw ise;
		}
		//tengo traccia dell'avvenuto cambiamento
		who = this.hashCode();
		when = System.currentTimeMillis();
		return true;
	}

	public boolean addAll(Collection<? extends E> c)
	{
		boolean retval = false;
		Iterator<? extends E> iter = c.iterator();
		while(iter.hasNext())
		{
			retval = retval || add(iter.next());
			/* Versione passo dopo passo
			Object tmp = iter.next();
			boolean tmpB = add(tmp);
			retval = retval || tmpB;
			*/
		}

		if(retval == true)
		{
			//tengo traccia dell'avvenuto cambiamento
			who = this.hashCode();
			when = System.currentTimeMillis();
		}
		return retval;
	}

	public Object[] toArray()
	{
		/*
		Object[] retval = new Object[size()];
		Iterator<E> iter = iterator();
		int i = 0;
		while(iter.hasNext())
		{
			retval[i++] = iter.next();
		}
		return retval;
		*/
		return toArray(new Object[0]);
	}

	public <T> T[] toArray(T[] a)
	{
		T[] retval = null;
		if(a.length >= size())
		{
			retval = a;
		}
		else
		{
			retval = (T[])new Object[size()];
		}
		Iterator iter = iterator();
		int i = 0;
		while(iter.hasNext())
		{
			retval[i++] = (T)iter.next();
		}
		return retval;
	}

	/* implementazione strongly tied all'implementazione corrente
	public Object[] toArray()
	{
		Object[] retval = new Object[count];
		for(int i = 0;i<count;i++)
		{
			retval[i] = storage[i];
		}
		return retval;
	}
	*/

	public void clear()
	{
		if(isEmpty())
		{
			return;
		}
		for(int i = 0;i<count;i++)
		{
			storage[i] = null;		
		}
		count = 0;
		//tengo traccia dell'avvenuto cambiamento
		who = this.hashCode();
		when = System.currentTimeMillis();
	}

	public boolean equals(Object o)
	{
		if((o instanceof Set) || (o instanceof List))
		{
			//da contratto nella documentazione: devo ritornare false
			return false;
		}

		/* Per un debug, ha senso voler ottenere l'eccezione
		if(!(o instanceof Collection))
		{
			return false;
		}
		*/

		Collection c = (Collection)o;
		//Vediamo se hanno la stessa lunghezza;
		if(c.size() != size())
		{
			return false;
		}

		//hanno la stessa lunghezza, mi tocca proprio confrontare
		// gli elementi a coppie.
		Iterator iter = c.iterator();
		Iterator<E> iterThis = iterator();
		Object tmp = null;
		while(iter.hasNext())
		{
			tmp = iter.next();
			if(tmp == null)
			{
				if(iterThis.next() != null)
				{
					return false;
				}
			}
			else
			{
				if(!tmp.equals(iterThis.next()))
				{
					return false;
				}
			}
		}
		return true;
	}

	public boolean isEmpty()
	{
		return count == 0;
	}

	public boolean removeAll(Collection<?> c)
	{
		boolean retval = false;
		Iterator iter = c.iterator();
		while(iter.hasNext())
		{
			retval = retval || remove(iter.next());
		}
		if(retval == true)
		{
			//tengo traccia dell'avvenuto cambiamento
			who = this.hashCode();
			when = System.currentTimeMillis();
		}
		return retval;
	}

	public boolean remove(Object o)
	{
		int i = 0;
		if(o == null)//cerco un null
		{
			for(i = 0;i < count;i++)
			{
				//se trovo esco dal loop
				if(storage[i] == null)
				{
					break;
				}
			}
		}
		else//uso equals per confrontare
		{
			for(i = 0;i < count;i++)
			{
				//se trovo esco dal loop
				if(o.equals(storage[i]))
				{
					break;
				}
			}
		}
		//se i vale meno di count, allora ho trovato l'elemento da eliminare
		//se i vale count, allora ho iterato su tutto il vettore senza successo
		if(i < count)
		{
			int k = 0;
			for(k = i; k < count-1;k++)
			{
				storage[k] = storage[k+1];
			}
			storage[k] = null;
			count--;

			//tengo traccia dell'avvenuto cambiamento
			who = this.hashCode();
			when = System.currentTimeMillis();

			return true;
		}
		return false;
	}

	/*
	public boolean remove(Object o)
	{
		if(o == null)
		{
			for(int i = 0;i < count;i++)
			{
				if(storage[i] == null)
				{
					for(int k = i; k < count-1;k++)
					{
						storage[k] = storage[k+1];
					}
					storage[k] = null;
					count--;
					return true;
				}
			}
		}
		else
		{
			for(int i = 0;i < count;i++)
			{
				if(o.equals(storage[i]))
				{
					for(int k = i; k < count-1;k++)
					{
						storage[k] = storage[k+1];
					}
					storage[k] = null;
					count--;
					return true;
				}
			}
		}
		return false;
	}
	*/

	public boolean contains(Object o)
	{
		if(o == null)
		{
			for(int i = 0;i<count;i++)
			{
				if(storage[i] == null)
				{
					return true;
				}
			}
		}
		else
		{
			for(int i = 0;i<count;i++)
			{
				if(o.equals(storage[i]))
				{
					return true;
				}
			}
		}
		return false;
	}

	public boolean containsAll(Collection<?> c)
	{
		Iterator iter = c.iterator();
		while(iter.hasNext())
		{
			if(!this.contains(iter.next()))
			{
				return false;
			}
		}
		return true;
	}

	public int size()
	{
		return count;
	}

	public boolean retainAll(Collection<?> c)
	{
		boolean retval = false;
		Iterator<E> iter = this.iterator();
		while(iter.hasNext())
		{
			E tmp = iter.next();
			if(c.contains(tmp))
			{
				//c'e', non faccio nulla
			}
			else
			{
				iter.remove();
				retval = true;
			}
		}
		if(retval == true)
		{
			//tengo traccia dell'avvenuto cambiamento
			who = this.hashCode();
			when = System.currentTimeMillis();
		}
		return retval;
	}

	public Iterator<E> iterator()
	{
		return new ArrayCollectionIterator<E>(this);
	}

	class ArrayCollectionIterator<E>
	implements Iterator<E>
	{
		long dob;
		boolean nextDone = false;
		int index;
		ArrayCollection ac;
		int me;

		ArrayCollectionIterator(ArrayCollection theCollection)
		{
			index = 0;
			ac = theCollection;
			dob = System.currentTimeMillis();
			me = this.hashCode();
		}

		private void isValid()
		{
			if(dob > ac.when)
			{
				return;
			}
			if(ac.who != me)
			{
				throw new IllegalStateException("Collection mutated and not by me!");
			}
		}

		public E next()
		{
			isValid();
			try
			{
				nextDone = true;
				return (E)ac.storage[index++];
			}
			catch(ArrayIndexOutOfBoundsException aioobe)
			{
				NoSuchElementException nsee = new NoSuchElementException();
				nsee.initCause(aioobe);
				throw nsee;
			}
		}

		public boolean hasNext()
		{
			isValid();
			return index < ac.count;
		}

		public void remove()
		{
			isValid();
			if(!nextDone)
			{
				throw new IllegalStateException();
			}
			nextDone = false;
			if(index == ac.count)
			{
				ac.storage[--index] = null;
				ac.count--;
			}
			else
			{
				int i = 0;
				for(i = index;i<ac.count-1;i++)
				{
					ac.storage[i-1] = ac.storage[i];
				}
				ac.storage[i] = null;
				ac.count--;
				index--;
			}
			
			//tengo traccia dell'ultima variazione
			ac.who = me;
			ac.when = System.currentTimeMillis();
		}
	}
}
